# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.


class OutcomeCache(object):
    """
    Base class for configuration outcome caching strategies. It does not
    implement a caching strategy itself (or, it implements the disable-cache
    strategy) but leaves the implementation of real strategies to subclasses.
    """

    def set_test_builder(self, test_builder):
        """
        Set the test builder for the cache.

        :param test_builder: Callable object that creates test case from a
            configuration. It must be identical to the test builder used by the
            tester class.
        """
        pass

    def add(self, config, result):
        """
        Add a new configuration to the cache.

        :param config: The configuration to save.
        :param result: The outcome of the added configuration.
        """
        pass

    def lookup(self, config):
        """
        Cache lookup to find out the outcome of a given configuration.

        :param config: The configuration we are looking for.
        :return: PASS or FAIL if config is in the cache; None, otherwise.
        """
        return None

    def clear(self):
        """
        Clear the cache.
        """
        pass

    def __str__(self):
        return '{}'


class ConfigCache(OutcomeCache):

    class _Entry(object):
        """
        This class holds test outcomes for configurations. This avoids running
        the same test twice.

        The outcome cache is implemented as a tree.  Each node points to the
        outcome of the remaining list.

        Example: ([1, 2, 3], PASS), ([1, 2], FAIL), ([1, 4, 5], FAIL):

             (2, FAIL)--(3, PASS)
            /
        (1, None)
            \
             (4, None)--(5, FAIL)
        """

        def __init__(self):
            self.result = None  # Result so far
            self.tail = {}  # Points to outcome of tail

    def __init__(self):
        self._root = self._Entry()

    def add(self, config, result):
        p = self._root
        for cs in config:
            if cs not in p.tail:
                p.tail[cs] = self._Entry()
            p = p.tail[cs]
        p.result = result

    def lookup(self, config):
        p = self._root
        for cs in config:
            if cs not in p.tail:
                return None
            p = p.tail[cs]
        return p.result

    def clear(self):
        self._root = self._Entry()

    def __str__(self):
        def _str(p):
            if p.result is not None:
                s.append('\t[%s]: %r,\n' % (', '.join(repr(cs) for cs in config), p.result.name))
            for cs, e in sorted(p.tail.items()):
                config.append(cs)
                _str(e)
                config.pop()

        config, s = [], []
        s.append('{\n')
        _str(self._root)
        s.append('}')
        return ''.join(s)


class ContentCache(OutcomeCache):
    """
    Class that can cache the outcome of test cases by their content.
    """

    def __init__(self):
        self.container = dict()
        self.test_builder = None

    def set_test_builder(self, test_builder):
        self.test_builder = test_builder

    def add(self, config, result):
        self.container[self.test_builder(config)] = result

    def lookup(self, config):
        return self.container.get(self.test_builder(config), None)

    def __str__(self):
        return '{\n%s}' % ''.join('\t%r: %r,\n' % (k, v.name) for k, v in sorted(self.container.items()))


# Aliases for cache classes to help their identification in CLI.
none = OutcomeCache
config = ConfigCache
content = ContentCache



# Copyright (c) 2016-2020 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.


class ZellerSplit(object):
    """
    Splits up the input config into n pieces as used by Zeller in the original
    reference implementation. The approach works iteratively in n steps, first
    slicing off a chunk sized 1/n-th of the original config, then slicing off
    1/(n-1)-th of the remainder, and so on, until the last piece is halved
    (always using integers division).
    """

    def __init__(self, n=2):
        """
        :param n: The split ratio used to determine how many parts (subsets) the
            config to split to (both initially and later on whenever config
            subsets needs to be re-split).
        """
        self._n = n

    def __call__(self, subsets):
        """
        :param subsets: List of sets that the current configuration is split to.
        :return: List of newly split sets.
        """
        config = [c for s in subsets for c in s]
        length = len(config)
        n = min(length, len(subsets) * self._n)

        next_subsets = []
        start = 0
        for i in range(n):
            stop = start + (length - start) // (n - i)
            next_subsets.append(config[start:stop])
            start = stop
        return next_subsets

    def __str__(self):
        cls = self.__class__
        return '%s.%s(n=%s)' % (cls.__module__, cls.__name__, self._n)


class BalancedSplit(object):
    """
    Slightly different version of Zeller's split. This version keeps the split
    balanced by distributing the residuals of the integer division among all
    chunks. This way, the size of the chunks in the resulting split is not
    monotonous.
    """

    def __init__(self, n=2):
        """
        :param n: The split ratio used to determine how many parts (subsets) the
            config to split to (both initially and later on whenever config
            subsets needs to be re-split).
        """
        self._n = n

    def __call__(self, subsets):
        """
        :param subsets: List of sets that the current configuration is split to.
        :return: List of newly split sets.
        """
        config = [c for s in subsets for c in s]
        length = len(config)
        n = min(length, len(subsets) * self._n)

        return [config[length * i // n:length * (i + 1) // n] for i in range(n)]

    def __str__(self):
        cls = self.__class__
        return '%s.%s(n=%s)' % (cls.__module__, cls.__name__, self._n)


# Aliases for split classes to help their identification in CLI.
zeller = ZellerSplit
balanced = BalancedSplit



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from . import parallel_loop
from .abstract_parallel_dd import AbstractParallelDD
from .config_iterators import forward
from .outcome import Outcome

logger = logging.getLogger(__name__)


class CombinedParallelDD(AbstractParallelDD):

    def __init__(self, test, *, split=None, cache=None, id_prefix=None,
                 proc_num=None, max_utilization=None,
                 config_iterator=None):
        """
        Initialize a CombinedParallelDD object.

        :param test: A callable tester object.
        :param split: Splitter method to break a configuration up to n part.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        :param proc_num: The level of parallelization.
        :param max_utilization: The maximum CPU utilization accepted.
        :param config_iterator: Reference to a generator function that provides
            config indices in an arbitrary order.
        """
        super().__init__(test=test, split=split, cache=cache, id_prefix=id_prefix, proc_num=proc_num, max_utilization=max_utilization)

        self._config_iterator = config_iterator or forward

    def _reduce_config(self, run, subsets, complement_offset):
        """
        Perform the reduce task using multiple processes. Subset and complement
        set tests are mixed and don't wait for each other.

        :param run: The index of the current iteration.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the
            index of the first unchecked complement (optimization purpose only).
        :return: Tuple: (list of subsets composing the failing config or None,
            next complement_offset).
        """
        n = len(subsets)
        self._fail_index.value = -1
        ploop = parallel_loop.Loop(self._proc_num, self._max_utilization)
        for i in self._config_iterator(2 * n):
            if i is None:
                continue

            if i < n:
                config_id = ('r%d' % run, 's%d' % i)
                config_set = subsets[i]
            else:
                i = int((i - n + complement_offset) % n) + n
                config_id = ('r%d' % run, 'c%d' % (i - n))
                config_set = [c for si, s in enumerate(subsets) for c in s if si != i - n]

            # If we checked this test before, return its result
            outcome = self._lookup_cache(config_set, config_id)
            if outcome is Outcome.PASS:
                continue
            if outcome is Outcome.FAIL:
                self._fail_index.value = i
                ploop.brk()
                break

            # Break if we found a FAIL either in the cache or be testing it now.
            if not ploop.do(self._loop_body, (config_set, i, config_id)):
                # if do() returned False, the test was not started
                break
        ploop.join()

        # fvalue contains the index of the cycle in the previous loop
        # which was found interesting. Otherwise it's -1.
        fvalue = self._fail_index.value
        if fvalue != -1:
            # Subset fail.
            if fvalue < n:
                return [subsets[fvalue]], 0
            # Complement fail.
            return subsets[:fvalue - n] + subsets[fvalue - n + 1:], fvalue - n

        return None, complement_offset



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging
import os
import signal
import sys

from multiprocessing import Array, Condition, cpu_count, Process, Value

import psutil

logger = logging.getLogger(__name__)
is_windows = sys.platform.startswith('win32')


# Beware! this is running in a new process now. state is shared with fork,
# but only changes to shared objects will be visible in parent.
def loop_body(shared_break, shared_slots, shared_lock, i, target, args):
    """
    Executes the given function in its own process group (on Windows: in its own
    process).

    :param shared_break: Loop-wide shared integer. Should be set to 1 if  the
        whole loop should be terminated.
    :param shared_slots: Loop-wide shared array of slots to keep track of status
        of loop bodies. The slot corresponding to a loop body should be set to 0
        once it has finished.
    :param shared_lock: Loop-wide shared lock. Should be notified when a loop
        body has finished.
    :param i: The index of the current slot.
    :param target: The function to run in parallel.
    :param args: The arguments that the target should run with.
    """
    if not is_windows:
        os.setpgrp()

    try:
        if not target(*args):
            shared_break.value = 1
    except Exception as e:
        logger.warning('', exc_info=e)
        shared_break.value = 1

    shared_slots[i] = 0

    with shared_lock:
        shared_lock.notify()


class Loop(object):
    """
    Parallel loop implementation based on multiprocessing module.
    """

    # TODO: could be None if we trusted that wait for infinity would not hang
    _timeout = 1

    def __init__(self, j=None, max_utilization=None):
        """
        Initialize a parallel loop object.

        :param j: The maximum number of parallel jobs.
        :param max_utilization: The maximum CPU utilization. Above this no more
            new jobs will be started.
        """
        self._j = j or cpu_count()
        self._max_utilization = max_utilization or 100
        # This gets initialized to 0, may be set to 1 anytime, but must not be reset to 0 ever;
        # thus, no locking is needed when accessing
        self._break = Value('i', 0, lock=False)
        self._lock = Condition()
        self._slots = Array('i', self._j, lock=False)
        self._procs = [None] * self._j
        psutil.cpu_percent(None)

    def _abort(self):
        """
        Terminate all live jobs.
        """
        for i, (slot, proc) in enumerate(zip(self._slots, self._procs)):
            if slot:
                if not is_windows:
                    try:
                        os.killpg(proc.pid, signal.SIGTERM)
                    except OSError:
                        # If the process with pid did not have time to become a process group leader,
                        # then pgid does not exist and os.killpg could not kill the process,
                        # so re-try kill the process only.
                        try:
                            os.kill(proc.pid, signal.SIGTERM)
                        except OSError:
                            pass
                else:
                    root_proc = psutil.Process(proc.pid)
                    children = root_proc.children(recursive=True) + [root_proc]
                    for child_proc in children:
                        try:
                            child_proc.terminate()
                        except psutil.Error:
                            pass
                    _, alive = psutil.wait_procs(children, timeout=1)
                    for proc in alive:
                        try:
                            proc.kill()
                        except psutil.Error:
                            pass
            self._slots[i], self._procs[i] = 0, None

    def _cleanup_slots(self):
        for i, (slot, proc) in enumerate(zip(self._slots, self._procs)):
            if slot:
                if not proc.is_alive() or not psutil.pid_exists(proc.pid):
                    self._slots[i], self._procs[i] = 0, None
            else:
                self._procs[i] = None

    # Target is expected to return True if loop shall continue and False if it
    # should break.
    def do(self, target, args):
        """
        Execute the target function if there is empty slot for it and the CPU
        utilization is suitable.

        :param target: The function to run in parallel.
        :param args: The parameters to run the target function with.
        :return: False if an interesting configuration was found. Otherwise
            returns True.
        """
        if self._break.value:
            logger.debug('do() called on a broken loop')
            return False

        i = None
        while True:
            if psutil.cpu_percent(None) <= self._max_utilization:
                self._cleanup_slots()
                i = next((x for x, slot in enumerate(self._slots) if not slot), None)
                if i is not None:
                    break
            with self._lock:
                self._lock.wait(self._timeout)

        if self._break.value:
            return False

        proc = Process(target=loop_body, args=(self._break, self._slots, self._lock, i, target, args))
        proc.start()
        self._slots[i], self._procs[i] = 1, proc

        return True

    def brk(self):
        """
        Break the parallel loop.
        """
        self._break.value = 1

    def join(self):
        """
        Wait until all the parallel child processes have finished.
        """
        if self._break.value:
            self._abort()
            return

        while True:
            self._cleanup_slots()
            # Return if all the jobs are done, that's there isn't any non-zero value in self._slots.
            if not any(self._slots):
                return

            with self._lock:
                self._lock.wait(self._timeout)

            if self._break.value:
                self._abort()



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from multiprocessing import Value

from .abstract_dd import AbstractDD
from .outcome import Outcome
from .outcome_cache import ConfigCache
from .shared_cache import shared_cache_decorator

logger = logging.getLogger(__name__)


class AbstractParallelDD(AbstractDD):
    """
    Abstract super-class of the various parallel DD implementations.
    """

    def __init__(self, test, *, split=None, cache=None, id_prefix=None,
                 proc_num=None, max_utilization=None):
        """
        Initialize an AbstractParallelDD object.

        :param test: A callable tester object.
        :param split: Splitter method to break a configuration up to n parts.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        :param proc_num: The level of parallelization.
        :param max_utilization: The maximum CPU utilization accepted.
        """
        cache = cache or shared_cache_decorator(ConfigCache)()
        super().__init__(test=test, split=split, cache=cache, id_prefix=id_prefix)

        self._proc_num = proc_num
        self._max_utilization = max_utilization
        self._fail_index = Value('i', -1, lock=False)

    def _loop_body(self, config, index, config_id):
        """
        The function that will be run in parallel.

        :param config: The list of entries of the current configuration.
        :param index: The index of the current configuration.
        :param config_id: The unique ID of the current configuration.
        :return: True if the test is not interesting, False otherwise.
        """
        if self._test_config(config, config_id) is Outcome.FAIL:
            self._fail_index.value = index
            return False

        return True



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging
import pytest

import picire


def interesting_a(c):
    if 7 in c and 2 not in c:
        return False
    if 5 in c and 8 in c:
        return True
    return False

config_a = [1, 2, 3, 4, 5, 6, 7, 8]
expect_a = [5, 8]


def interesting_b(c):
    if not c:
        return False
    if 1 in c and 2 in c and 3 in c and 4 in c and \
                    5 in c and 6 in c and 7 in c and 8 in c:
        return True
    return False

config_b = [1, 2, 3, 4, 5, 6, 7, 8]
expect_b = [1, 2, 3, 4, 5, 6, 7, 8]


def interesting_c(c):
    if 1 in c and 2 in c and 3 in c and 4 in c and \
       6 in c and 8 in c:
        return True
    return False

config_c = [1, 2, 3, 4, 5, 6, 7, 8]
expect_c = [1, 2, 3, 4, 6, 8]


class CaseTest:

    def __init__(self, interesting, content):
        self.content = content
        self.interesting = interesting

    def __call__(self, config, config_id):
        return picire.Outcome.FAIL if self.interesting([self.content[x] for x in config]) else picire.Outcome.PASS


@pytest.mark.parametrize('interesting, config, expect', [
    (interesting_a, config_a, expect_a),
    (interesting_b, config_b, expect_b),
    (interesting_c, config_c, expect_c),
])
@pytest.mark.parametrize('granularity', [
    2,
    float('inf'),
])
class TestApi:

    def _run_picire(self, interesting, config, expect, granularity, dd, split, subset_first, subset_iterator, complement_iterator, cache):
        if dd != picire.CombinedParallelDD:
            it_kwargs = {
                'subset_first': subset_first,
                'subset_iterator': subset_iterator,
                'complement_iterator': complement_iterator,
            }
        else:
            it_kwargs = {
                'config_iterator': picire.CombinedIterator(subset_first, subset_iterator, complement_iterator)
            }

        if dd != picire.DD:
            cache = picire.shared_cache_decorator(cache)

        logging.basicConfig(format='%(message)s')
        logging.getLogger('picire').setLevel(logging.DEBUG)

        dd_obj = dd(CaseTest(interesting, config),
                    split=split(n=granularity),
                    cache=cache(),
                    **it_kwargs)
        output = [config[x] for x in dd_obj(list(range(len(config))))]

        assert output == expect

    @pytest.mark.parametrize('split, subset_first, subset_iterator, complement_iterator, cache', [
        (picire.config_splitters.balanced, True, picire.config_iterators.forward, picire.config_iterators.forward, picire.OutcomeCache),
        (picire.config_splitters.zeller, True, picire.config_iterators.forward, picire.config_iterators.backward, picire.ConfigCache),
        (picire.config_splitters.balanced, False, picire.config_iterators.backward, picire.config_iterators.forward, picire.OutcomeCache),
        (picire.config_splitters.zeller, False, picire.config_iterators.backward, picire.config_iterators.backward, picire.ConfigCache),
        (picire.config_splitters.balanced, True, picire.config_iterators.skip, picire.config_iterators.forward, picire.OutcomeCache),
        (picire.config_splitters.zeller, True, picire.config_iterators.skip, picire.config_iterators.backward, picire.ConfigCache),
    ])
    def test_dd(self, interesting, config, expect, granularity, split, subset_first, subset_iterator, complement_iterator, cache):
        self._run_picire(interesting, config, expect, granularity, picire.DD, split, subset_first, subset_iterator, complement_iterator, cache)

    @pytest.mark.parametrize('split, subset_first, subset_iterator, complement_iterator, cache', [
        (picire.config_splitters.zeller, False, picire.config_iterators.forward, picire.config_iterators.forward, picire.ConfigCache),
        (picire.config_splitters.balanced, False, picire.config_iterators.forward, picire.config_iterators.backward, picire.OutcomeCache),
        (picire.config_splitters.zeller, True, picire.config_iterators.backward, picire.config_iterators.forward, picire.ConfigCache),
        (picire.config_splitters.balanced, True, picire.config_iterators.backward, picire.config_iterators.backward, picire.OutcomeCache),
        (picire.config_splitters.zeller, True, picire.config_iterators.skip, picire.config_iterators.forward, picire.ConfigCache),
        (picire.config_splitters.balanced, True, picire.config_iterators.skip, picire.config_iterators.backward, picire.OutcomeCache),
    ])
    def test_parallel(self, interesting, config, expect, granularity, split, subset_first, subset_iterator, complement_iterator, cache):
        self._run_picire(interesting, config, expect, granularity, picire.ParallelDD, split, subset_first, subset_iterator, complement_iterator, cache)

    @pytest.mark.parametrize('split, subset_first, subset_iterator, complement_iterator, cache', [
        (picire.config_splitters.zeller, True, picire.config_iterators.forward, picire.config_iterators.forward, picire.OutcomeCache),
        (picire.config_splitters.balanced, True, picire.config_iterators.forward, picire.config_iterators.backward, picire.ConfigCache),
        (picire.config_splitters.zeller, False, picire.config_iterators.backward, picire.config_iterators.forward, picire.OutcomeCache),
        (picire.config_splitters.balanced, False, picire.config_iterators.backward, picire.config_iterators.backward, picire.ConfigCache),
    ])
    def test_combined(self, interesting, config, expect, granularity, split, subset_first, subset_iterator, complement_iterator, cache):
        self._run_picire(interesting, config, expect, granularity, picire.CombinedParallelDD, split, subset_first, subset_iterator, complement_iterator, cache)



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import itertools
import logging

from .config_splitters import ZellerSplit
from .outcome import Outcome
from .outcome_cache import OutcomeCache

logger = logging.getLogger(__name__)


class AbstractDD(object):
    """
    Abstract super-class of the parallel and non-parallel DD classes.
    """

    def __init__(self, test, *, split=None, cache=None, id_prefix=None):
        """
        Initialise an abstract DD class. Not to be called directly, only by
        super calls in subclass initializers.

        :param test: A callable tester object.
        :param split: Splitter method to break a configuration up to n parts.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        """
        self._test = test
        self._split = split or ZellerSplit()
        self._cache = cache or OutcomeCache()
        self._id_prefix = id_prefix or ()

    def __call__(self, config):
        """
        Return a 1-minimal failing subset of the initial configuration.

        :param config: The initial configuration that will be reduced.
        :return: 1-minimal failing configuration.
        """
        subsets = [config]
        complement_offset = 0

        for run in itertools.count():
            logger.info('Run #%d', run)
            logger.info('\tConfig size: %d', len(config))
            assert self._test_config(config, ('r%d' % run, 'assert')) is Outcome.FAIL

            # Minimization ends if the configuration is already reduced to a single unit.
            if len(config) < 2:
                logger.info('\tGranularity: %d', len(subsets))
                logger.debug('\tConfig: %r', subsets)
                logger.info('\tDone')
                return config

            if len(subsets) < 2:
                assert len(subsets) == 1
                subsets = self._split(subsets)

            logger.info('\tGranularity: %d', len(subsets))
            logger.debug('\tConfig: %r', subsets)

            next_subsets, complement_offset = self._reduce_config(run, subsets, complement_offset)

            if next_subsets is not None:
                # Interesting configuration is found, start new iteration.
                subsets = next_subsets
                config = [c for s in subsets for c in s]

                logger.info('\tReduced')

            elif len(subsets) < len(config):
                # No interesting configuration is found but it is still not the finest splitting, start new iteration.
                next_subsets = self._split(subsets)
                complement_offset = (complement_offset * len(next_subsets)) / len(subsets)
                subsets = next_subsets

                logger.info('\tIncreased granularity')

            else:
                # Minimization ends if no interesting configuration was found by the finest splitting.
                logger.info('\tDone')
                return config

    def _reduce_config(self, run, subsets, complement_offset):
        """
        Perform the reduce task of ddmin. To be overridden by subclasses.

        :param run: The index of the current iteration.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the
            index of the first unchecked complement (optimization purpose only).
        :return: Tuple: (list of subsets composing the failing config or None,
            next complement_offset).
        """
        raise NotImplementedError()

    def _lookup_cache(self, config, config_id):
        """
        Perform a cache lookup if caching is enabled.

        :param config: The configuration we are looking for.
        :param config_id: The ID describing the configuration (only for debug
            message).
        :return: None if outcome is not found for config in cache or if caching
            is disabled, PASS or FAIL otherwise.
        """
        cached_result = self._cache.lookup(config)
        if cached_result is not None:
            logger.debug('\t[ %s ]: cache = %r', self._pretty_config_id(self._id_prefix + config_id), cached_result.name)

        return cached_result

    def _test_config(self, config, config_id):
        """
        Test a single configuration and save the result in cache.

        :param config: The current configuration to test.
        :param config_id: Unique ID that will be used to save tests to easily
            identifiable directories.
        :return: PASS or FAIL
        """
        config_id = self._id_prefix + config_id

        logger.debug('\t[ %s ]: test...', self._pretty_config_id(config_id))
        outcome = self._test(config, config_id)
        logger.debug('\t[ %s ]: test = %r', self._pretty_config_id(config_id), outcome.name)

        if 'assert' not in config_id:
            self._cache.add(config, outcome)

        return outcome

    @staticmethod
    def _pretty_config_id(config_id):
        """
        Create beautified identifier for the current task from the argument.
        The argument is typically a tuple in the form of ('rN', 'DM'), where N
        is the index of the current iteration, D is direction of reduce (either
        s(ubset) or c(omplement)), and M is the index of the current test in the
        iteration. Alternatively, argument can also be in the form of
        (rN, 'assert') for double checking the input at the start of an
        iteration.

        :param config_id: Config ID tuple.
        :return: Concatenating the arguments with slashes, e.g., "rN / DM".
        """
        return ' / '.join(str(i) for i in config_id)



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from .abstract_dd import AbstractDD
from .config_iterators import forward
from .outcome import Outcome
from .outcome_cache import ConfigCache

logger = logging.getLogger(__name__)


class DD(AbstractDD):
    """
    Single process version of the Delta Debugging algorithm.
    """

    def __init__(self, test, *, split=None, cache=None, id_prefix=None,
                 subset_first=True, subset_iterator=None, complement_iterator=None):
        """
        Initialize a DD object.

        :param test: A callable tester object.
        :param split: Splitter method to break a configuration up to n parts.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        :param subset_first: Boolean value denoting whether the reduce has to
            start with the subset-based approach or not.
        :param subset_iterator: Reference to a generator function that provides
            config indices in an arbitrary order.
        :param complement_iterator: Reference to a generator function that
            provides config indices in an arbitrary order.
        """
        cache = cache or ConfigCache()
        super().__init__(test=test, split=split, cache=cache, id_prefix=id_prefix)

        self._subset_iterator = subset_iterator or forward
        self._complement_iterator = complement_iterator or forward

        if subset_first:
            self._first_reduce, self._second_reduce = self._reduce_to_subset, self._reduce_to_complement
        else:
            self._first_reduce, self._second_reduce = self._reduce_to_complement, self._reduce_to_subset

    def _reduce_config(self, run, subsets, complement_offset):
        """
        Perform the reduce task in single process mode.

        :param run: The index of the current iteration.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the
            index of the first unchecked complement (optimization purpose only).
        :return: Tuple: (list of subsets composing the failing config or None,
            next complement_offset).
        """
        next_subsets, complement_offset = self._first_reduce(run, subsets, complement_offset)
        if next_subsets is None:
            next_subsets, complement_offset = self._second_reduce(run, subsets, complement_offset)

        return next_subsets, complement_offset

    def _reduce_to_subset(self, run, subsets, complement_offset):
        """
        Perform a subset-based reduce task.

        :param run: The index of the current iteration.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the
            index of the first unchecked complement (optimization purpose only).
        :return: Tuple: (list of subsets composing the failing config or None,
            next complement_offset).
        """
        n = len(subsets)
        for i in self._subset_iterator(n):
            if i is None:
                continue

            config_id = ('r%d' % run, 's%d' % i)
            subset = subsets[i]

            # Get the outcome either from cache or by testing it.
            outcome = self._lookup_cache(subset, config_id) or self._test_config(subset, config_id)
            if outcome is Outcome.FAIL:
                # Interesting subset is found.
                return [subsets[i]], 0

        return None, complement_offset

    def _reduce_to_complement(self, run, subsets, complement_offset):
        """
        Perform a complement-based reduce task.

        :param run: The index of the current iteration.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the
            index of the first unchecked complement (optimization purpose only).
        :return: Tuple: (list of subsets composing the failing config or None,
            next complement_offset).
        """
        n = len(subsets)
        for i in self._complement_iterator(n):
            if i is None:
                continue
            i = int((i + complement_offset) % n)

            config_id = ('r%d' % run, 'c%d' % i)
            complement = [c for si, s in enumerate(subsets) for c in s if si != i]

            outcome = self._lookup_cache(complement, config_id) or self._test_config(complement, config_id)
            if outcome is Outcome.FAIL:
                # Interesting complement is found.
                # In next run, start removing the following subset
                return subsets[:i] + subsets[i + 1:], i

        return None, complement_offset



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from . import parallel_loop
from .abstract_parallel_dd import AbstractParallelDD
from .config_iterators import forward
from .outcome import Outcome

logger = logging.getLogger(__name__)


class ParallelDD(AbstractParallelDD):

    def __init__(self, test, *, split=None, cache=None, id_prefix=None,
                 proc_num=None, max_utilization=None,
                 subset_first=True, subset_iterator=None, complement_iterator=None):
        """
        Initialize a ParallelDD object.

        :param test: A callable tester object.
        :param split: Splitter method to break a configuration up to n part.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        :param proc_num: The level of parallelization.
        :param max_utilization: The maximum CPU utilization accepted.
        :param subset_first: Boolean value denoting whether the reduce has to
            start with the subset-based approach or not.
        :param subset_iterator: Reference to a generator function that provides
            config indices in an arbitrary order.
        :param complement_iterator: Reference to a generator function that
            provides config indices in an arbitrary order.
        """
        super().__init__(test=test, split=split, cache=cache, id_prefix=id_prefix, proc_num=proc_num, max_utilization=max_utilization)

        self._subset_iterator = subset_iterator or forward
        self._complement_iterator = complement_iterator or forward

        if subset_first:
            self._first_reduce, self._second_reduce = self._reduce_to_subset, self._reduce_to_complement
        else:
            self._first_reduce, self._second_reduce = self._reduce_to_complement, self._reduce_to_subset

    def _reduce_config(self, run, subsets, complement_offset):
        """
        Perform the reduce task using multiple processes.
        Subset and complement set tests are executed sequentially.

        :param run: The index of the current iteration.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the
            index of the first unchecked complement (optimization purpose only).
        :return: Tuple: (list of subsets composing the failing config or None,
            next complement_offset).
        """
        next_subsets, complement_offset = self._first_reduce(run, subsets, complement_offset)
        if next_subsets is None:
            next_subsets, complement_offset = self._second_reduce(run, subsets, complement_offset)

        return next_subsets, complement_offset

    def _reduce_to_subset(self, run, subsets, complement_offset):
        """
        Perform a subset-based reduce task.

        :param run: The index of the current iteration.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the
            index of the first unchecked complement (optimization purpose only).
        :return: Tuple: (list of subsets composing the failing config or None,
            next complement_offset).
        """
        # Looping through the subsets.
        n = len(subsets)
        self._fail_index.value = -1
        ploop = parallel_loop.Loop(self._proc_num, self._max_utilization)
        for i in self._subset_iterator(n):
            if i is None:
                continue

            config_id = ('r%d' % run, 's%d' % i)
            subset = subsets[i]

            # If we had this test before, return the saved result.
            outcome = self._lookup_cache(subset, config_id)
            if outcome is Outcome.PASS:
                continue
            if outcome is Outcome.FAIL:
                self._fail_index.value = i
                ploop.brk()
                break

            if not ploop.do(self._loop_body, (subset, i, config_id)):
                # if do() returned False, the test was not started
                break
        ploop.join()

        fvalue = self._fail_index.value
        if fvalue != -1:
            return [subsets[fvalue]], 0

        return None, complement_offset

    def _reduce_to_complement(self, run, subsets, complement_offset):
        """
        Perform a complement-based reduce task.

        :param run: The index of the current iteration.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the
            index of the first unchecked complement (optimization purpose only).
        :return: Tuple: (list of subsets composing the failing config or None,
            next complement_offset).
        """
        n = len(subsets)
        self._fail_index.value = -1
        ploop = parallel_loop.Loop(self._proc_num, self._max_utilization)
        for i in self._complement_iterator(n):
            if i is None:
                continue
            i = int((i + complement_offset) % n)

            config_id = ('r%d' % run, 'c%d' % i)
            complement = [c for si, s in enumerate(subsets) for c in s if si != i]

            # If we had this test before, return its result
            outcome = self._lookup_cache(complement, config_id)
            if outcome is Outcome.PASS:
                continue
            if outcome is Outcome.FAIL:
                self._fail_index.value = i
                ploop.brk()
                break

            if not ploop.do(self._loop_body, (complement, i, config_id)):
                # If do() returned False, the test was not started.
                break
        ploop.join()

        fvalue = self._fail_index.value
        if fvalue != -1:
            # In next run, start removing the following subset.
            return subsets[:fvalue] + subsets[fvalue + 1:], fvalue

        return None, complement_offset



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import os
import platform
import pytest
import subprocess
import sys


is_windows = sys.platform.startswith('win32')
script_ext = '.bat' if is_windows else '.sh'

is_cpython = platform.python_implementation() == 'CPython'

tests_dir = os.path.dirname(os.path.abspath(__file__))
resources_dir = os.path.join(tests_dir, 'resources')


@pytest.mark.parametrize('test, inp, exp, args_atom', [
    pytest.param('test-json-extra-comma', 'inp-extra-comma.json', 'exp-extra-comma.json', ('--atom=line', ),
                 marks=pytest.mark.skipif(not is_cpython, reason='json error messages are implementation-specific')),
    ('test-sumprod10-sum', 'inp-sumprod10.py', 'exp-sumprod10-sum.py', ('--atom=line', )),
    ('test-sumprod10-prod', 'inp-sumprod10.py', 'exp-sumprod10-prod.py', ('--atom=line', )),
    pytest.param('test-json-invalid-escape', 'inp-invalid-escape.json', 'exp-invalid-escape.json', ('--atom=char', ),
                 marks=pytest.mark.skipif(not is_cpython, reason='json error messages are implementation-specific')),
    pytest.param('test-json-invalid-escape', 'inp-invalid-escape.json', 'exp-invalid-escape.json', ('--atom=both', ),
                 marks=pytest.mark.skipif(not is_cpython, reason='json error messages are implementation-specific')),
])
class TestCli:

    def _run_picire(self, test, inp, exp, tmpdir, args):
        out_dir = str(tmpdir)
        cmd = (sys.executable, '-m', 'picire') \
              + ('--test=' + test + script_ext, '--input=' + inp, '--out=' + out_dir) \
              + ('--log-level=TRACE', ) \
              + args
        subprocess.run(cmd, cwd=resources_dir, check=True)

        with open(os.path.join(out_dir, inp), 'rb') as outf:
            outb = outf.read()
        with open(os.path.join(resources_dir, exp), 'rb') as expf:
            expb = expf.read()
        assert outb == expb

    @pytest.mark.parametrize('args', [
        ('--split=balanced', '--subset-iterator=forward', '--complement-iterator=forward', '--cache=none'),
        ('--split=zeller', '--subset-iterator=forward', '--complement-iterator=backward', '--cache=config'),
        ('--split=balanced', '--complement-first', '--subset-iterator=backward', '--complement-iterator=forward', '--cache=content'),
        ('--split=zeller', '--complement-first', '--subset-iterator=backward', '--complement-iterator=backward', '--cache=none'),
        ('--split=balanced', '--subset-iterator=skip', '--complement-iterator=forward', '--cache=config'),
        ('--split=zeller', '--subset-iterator=skip', '--complement-iterator=backward', '--cache=content'),
    ])
    def test_dd(self, test, inp, exp, tmpdir, args_atom, args):
        self._run_picire(test, inp, exp, tmpdir, args_atom + args)

    @pytest.mark.parametrize('args', [
        ('--split=zeller', '--complement-first', '--subset-iterator=forward', '--complement-iterator=forward', '--cache=config'),
        ('--split=balanced', '--complement-first', '--subset-iterator=forward', '--complement-iterator=backward', '--cache=content'),
        ('--split=zeller', '--subset-iterator=backward', '--complement-iterator=forward', '--cache=none'),
        ('--split=balanced', '--subset-iterator=backward', '--complement-iterator=backward', '--cache=config'),
        ('--split=zeller', '--subset-iterator=skip', '--complement-iterator=forward', '--cache=content'),
        ('--split=balanced', '--subset-iterator=skip', '--complement-iterator=backward', '--cache=none'),
    ])
    def test_parallel(self, test, inp, exp, tmpdir, args_atom, args):
        self._run_picire(test, inp, exp, tmpdir, args_atom + ('--parallel', ) + args)

    @pytest.mark.parametrize('args', [
        ('--split=zeller', '--subset-iterator=forward', '--complement-iterator=forward', '--cache=content'),
        ('--split=balanced', '--subset-iterator=forward', '--complement-iterator=backward', '--cache=none'),
        ('--split=zeller', '--complement-first', '--subset-iterator=backward', '--complement-iterator=forward', '--cache=config'),
        ('--split=balanced', '--complement-first', '--subset-iterator=backward', '--complement-iterator=backward', '--cache=content'),
    ])
    def test_combined(self, test, inp, exp, tmpdir, args_atom, args):
        self._run_picire(test, inp, exp, tmpdir, args_atom + ('--parallel', '--combine-loops', ) + args)



# Copyright (c) 2016-2019 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from . import config_iterators


class CombinedIterator(object):
    """
    Callable iterator class that acts as generator when subset and complement
    check loops are combined.
    """

    def __init__(self, subset_first=True, subset_iterator=config_iterators.forward, complement_iterator=config_iterators.forward):
        """
        Initialize CombinedIterator with the basic settings describing the order
        of subset and complement checks together with the certain iterators.

        :param subset_first: Boolean value denoting whether the reduce has to
            start with the subset based approach or not.
        :param subset_iterator: Reference to a generator function that provides
            config indices in an arbitrary order.
        :param complement_iterator: Reference to a generator function that
            provides config indices in an arbitrary order.
        """
        self._subset_first = subset_first
        self._subset_iterator = subset_iterator
        self._complement_iterator = complement_iterator

    def __call__(self, n):
        """
        Provide the index of the next configuration according to the settings.

        :param n: The number of configurations (subset and combined configs
            altogether).
        :return: The index of the next configuration.
        """
        if self._subset_first:
            for i in self._subset_iterator(n // 2):
                yield i
            for i in self._complement_iterator(n // 2):
                yield i + n // 2
        else:
            for i in self._complement_iterator(n // 2):
                yield i + n // 2
            for i in self._subset_iterator(n // 2):
                yield i



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import inspect

from multiprocessing import Lock
from multiprocessing.managers import BaseManager

from .outcome_cache import OutcomeCache


class SharedCacheManager(BaseManager):
    """
    Manager to share the cache object between parallel processes.
    """

    _managers = {}


class SharedCacheConstructor(object):
    """
    Wrapper for cache classes to instantiate them as shared.
    """

    _exposed = []

    @classmethod
    def exposed(cls, fn):
        cls._exposed.append(fn.__name__)
        return fn

    def __init__(self, cache_class):
        self._cache_class = cache_class

    def __call__(self, *args, **kwargs):
        return SharedCache(self._cache_class(*args, **kwargs))


class SharedCache(OutcomeCache):
    """
    Thread-safe cache representation that stores the evaluated configurations
    and their outcome.
    """

    def __init__(self, cache):
        self._cache = cache
        self._lock = Lock()

    @SharedCacheConstructor.exposed
    def set_test_builder(self, test_builder):
        with self._lock:
            self._cache.set_test_builder(test_builder)

    @SharedCacheConstructor.exposed
    def add(self, config, result):
        with self._lock:
            self._cache.add(config, result)

    @SharedCacheConstructor.exposed
    def lookup(self, config):
        with self._lock:
            return self._cache.lookup(config)

    @SharedCacheConstructor.exposed
    def clear(self):
        with self._lock:
            self._cache.clear()

    @SharedCacheConstructor.exposed
    def __str__(self):
        with self._lock:
            return self._cache.__str__()


def shared_cache_decorator(cache_class):
    name = cache_class.__name__

    if not hasattr(SharedCacheManager, name):
        SharedCacheManager.register(name, SharedCacheConstructor(cache_class), None, SharedCacheConstructor._exposed)
        getattr(SharedCacheManager, name).__signature__ = inspect.signature(cache_class)

        SharedCacheManager._managers[name] = SharedCacheManager()
        SharedCacheManager._managers[name].start()

    return getattr(SharedCacheManager._managers[name], name)



# Copyright (c) 2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from enum import Enum


class Outcome(Enum):

    PASS = 'PASS'
    FAIL = 'FAIL'

    def __repr__(self):
        return '<%s.%s>' % (self.__class__.__name__, self.name)



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from . import cli
from . import config_iterators
from . import config_splitters
from .abstract_dd import AbstractDD
from .abstract_parallel_dd import AbstractParallelDD
from .cli import __version__, reduce
from .combined_iterator import CombinedIterator
from .combined_parallel_dd import CombinedParallelDD
from .dd import DD
from .outcome import Outcome
from .outcome_cache import ConfigCache, ContentCache, OutcomeCache
from .parallel_dd import ParallelDD
from .shared_cache import shared_cache_decorator
from .subprocess_test import ConcatTestBuilder, SubprocessTest



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import argparse
import codecs
import os
import time

from multiprocessing import cpu_count
from os.path import basename, exists, join, realpath
from shutil import rmtree

import chardet
import inators
import pkg_resources

from inators import log as logging

from . import config_iterators, config_splitters, outcome_cache
from .combined_iterator import CombinedIterator
from .combined_parallel_dd import CombinedParallelDD
from .dd import DD
from .parallel_dd import ParallelDD
from .shared_cache import shared_cache_decorator
from .subprocess_test import ConcatTestBuilder, SubprocessTest

logger = logging.getLogger('picire')
__version__ = pkg_resources.get_distribution(__package__).version


def create_parser():
    def int_or_inf(value):
        if value == 'inf':
            return float('inf')
        value = int(value)
        if value < 2:
            raise argparse.ArgumentTypeError('invalid value: {value!r} (must be at least 2)'.format(value=value))
        return value

    parser = argparse.ArgumentParser(description='Command line interface of the "picire" test case reducer')
    parser.add_argument('-i', '--input', metavar='FILE', required=True,
                        help='test case to be reduced')

    # Base reduce settings.
    parser.add_argument('--cache', metavar='NAME',
                        choices=[i for i in dir(outcome_cache) if not i.startswith('_') and i.islower()], default='config',
                        help='cache strategy (%(choices)s; default: %(default)s)')
    parser.add_argument('--split', metavar='NAME',
                        choices=[i for i in dir(config_splitters) if not i.startswith('_') and i.islower()], default='zeller',
                        help='split algorithm (%(choices)s; default: %(default)s)')
    parser.add_argument('--test', metavar='FILE', required=True,
                        help='test command that decides about interestingness of an input')
    parser.add_argument('--granularity', metavar='N', type=int_or_inf, default=2,
                        help='initial granularity and split factor (integer or \'inf\'; default: %(default)d)')
    parser.add_argument('--encoding', metavar='NAME',
                        help='test case encoding (default: autodetect)')

    # Extra settings for parallel reduce.
    parser.add_argument('-p', '--parallel', action='store_true', default=False,
                        help='run DD in parallel')
    parser.add_argument('-c', '--combine-loops', action='store_true', default=False,
                        help='combine subset and complement check loops for more parallelization (has effect in parallel mode only)')
    parser.add_argument('-j', '--jobs', metavar='N', type=int, default=cpu_count(),
                        help='maximum number of test commands to execute in parallel (has effect in parallel mode only; default: %(default)s)')
    parser.add_argument('-u', '--max-utilization', metavar='N', type=int, default=100,
                        help='maximum CPU utilization allowed; don\'t start new parallel jobs until utilization is higher (has effect in parallel mode only; default: %(default)s)')

    # Tweaks how to walk through the chunk lists.
    parser.add_argument('--complement-first', dest='subset_first', action='store_false', default=True,
                        help='check complements first')
    parser.add_argument('--subset-iterator', metavar='NAME',
                        choices=[i for i in dir(config_iterators) if not i.startswith('_')], default='forward',
                        help='ordering strategy for looping through subsets (%(choices)s; default: %(default)s)')
    parser.add_argument('--complement-iterator', metavar='NAME',
                        choices=[i for i in dir(config_iterators) if not i.startswith('_')], default='forward',
                        help='ordering strategy for looping through complements (%(choices)s; default: %(default)s)')

    # Additional settings.
    inators.arg.add_log_level_argument(parser)
    parser.add_argument('-o', '--out', metavar='DIR',
                        help='working directory (default: input.timestamp)')
    parser.add_argument('--no-cleanup', dest='cleanup', default=True, action='store_false',
                        help='disable the removal of generated temporary files')
    return parser


def process_args(args):
    inators.arg.process_log_level_argument(args, logger)

    args.input = realpath(args.input)
    if not exists(args.input):
        raise ValueError('Test case does not exist: %s' % args.input)

    with open(args.input, 'rb') as f:
        args.src = f.read()

    if args.encoding:
        try:
            codecs.lookup(args.encoding)
        except LookupError as e:
            raise ValueError('The given encoding (%s) is not known.' % args.encoding) from e
    else:
        args.encoding = chardet.detect(args.src)['encoding'] or 'latin-1'

    args.src = args.src.decode(args.encoding)

    args.out = realpath(args.out if args.out else '%s.%s' % (args.input, time.strftime('%Y%m%d_%H%M%S')))

    args.test = realpath(args.test)
    if not exists(args.test) or not os.access(args.test, os.X_OK):
        raise ValueError('Tester program does not exist or isn\'t executable: %s' % args.test)

    args.tester_class = SubprocessTest
    args.tester_config = dict(command_pattern=[args.test, '%s'],
                              work_dir=join(args.out, 'tests'),
                              filename=basename(args.input),
                              encoding=args.encoding,
                              cleanup=args.cleanup)

    args.cache = getattr(outcome_cache, args.cache)
    if args.parallel:
        args.cache = shared_cache_decorator(args.cache)

    split_class = getattr(config_splitters, args.split)
    subset_iterator = getattr(config_iterators, args.subset_iterator)
    complement_iterator = getattr(config_iterators, args.complement_iterator)

    # Choose the reducer class that will be used and its configuration.
    if not args.parallel:
        args.reduce_class = DD
        args.reduce_config = dict(subset_iterator=subset_iterator,
                                  complement_iterator=complement_iterator,
                                  subset_first=args.subset_first)
    else:
        if args.combine_loops:
            args.reduce_class = CombinedParallelDD
            args.reduce_config = dict(config_iterator=CombinedIterator(args.subset_first, subset_iterator, complement_iterator))
        else:
            args.reduce_class = ParallelDD
            args.reduce_config = dict(subset_iterator=subset_iterator,
                                      complement_iterator=complement_iterator,
                                      subset_first=args.subset_first)
        args.reduce_config.update(dict(proc_num=args.jobs,
                                       max_utilization=args.max_utilization))
    args.reduce_config.update(dict(split=split_class(n=args.granularity)))

    logger.info('Input loaded from %s', args.input)


def log_args(title, args):
    def _log_args(args):
        if not args:
            return repr(args)
        if isinstance(args, dict):
            log = []
            for k, v in sorted(args.items()):
                k_log = _log_args(k)
                v_log = _log_args(v)
                if isinstance(v_log, list):
                    log += ['%s:' % k_log]
                    for line in v_log:
                        log += ['\t' + line]
                else:
                    log += ['%s: %s' % (k_log, v_log)]
            return log if len(log) > 1 else log[0]
        if isinstance(args, list):
            v_logs = [_log_args(v) for v in args]
            if any(isinstance(v_log, list) for v_log in v_logs):
                log = []
                for v_log in v_logs:
                    if not isinstance(v_log, list):
                        v_log = [v_log]
                    for i, line in enumerate(v_log):
                        log += ['%s %s' % ('-' if i == 0 else ' ', line)]
            else:
                log = ', '.join(v_log for v_log in v_logs)
            return log
        if hasattr(args, '__name__'):
            return '.'.join(([args.__module__] if hasattr(args, '__module__') else []) + [args.__name__])
        return str(args)
    logger.info('%s\n\t%s\n', title, '\n\t'.join(_log_args(args)))


def reduce(src, *,
           reduce_class, reduce_config,
           tester_class, tester_config,
           atom='line',
           cache_class=None):
    """
    Execute picire as if invoked from command line, however, control its
    behaviour not via command line arguments but function parameters.

    :param src: Contents of the test case to reduce.
    :param reduce_class: Reference to the reducer class.
    :param reduce_config: Dictionary containing information to initialize the
        reduce_class.
    :param tester_class: Reference to a runnable class that can decide about the
        interestingness of a test case.
    :param tester_config: Dictionary containing information to initialize the
        tester_class.
    :param atom: Input granularity to work with during reduce ('char', 'line',
        or 'both'; default: 'line').
    :param cache_class: Reference to the cache class to use.
    :return: The contents of the minimal test case.
    """

    # Get the parameters in a dictionary so that they can be pretty-printed
    # (minus src, as that parameter can be arbitrarily large)
    args = locals().copy()
    del args['src']
    log_args('Reduce session starts', args)

    cache = cache_class() if cache_class else None

    for atom_cnt, atom_name in enumerate(['line', 'char'] if atom == 'both' else [atom]):
        # Split source to the chosen atoms.
        if atom_name == 'line':
            src = src.splitlines(True)
        logger.info('Initial test contains %d %ss', len(src), atom_name)

        test_builder = ConcatTestBuilder(src)
        if cache:
            cache.clear()
            cache.set_test_builder(test_builder)

        dd = reduce_class(tester_class(test_builder=test_builder, **tester_config),
                          cache=cache,
                          id_prefix=('a%d' % atom_cnt,),
                          **reduce_config)
        min_set = dd(list(range(len(src))))
        src = test_builder(min_set)

        logger.trace('The cached results are: %s', cache)
        logger.debug('A minimal config is: %r', min_set)

    return src


def postprocess(args, out_src):
    if args.cleanup:
        rmtree(join(args.out, 'tests'))

    output = join(args.out, basename(args.input))
    with codecs.open(output, 'w', encoding=args.encoding, errors='ignore') as f:
        f.write(out_src)

    logger.info('Output saved to %s', output)


def execute():
    """
    The main entry point of picire.
    """
    logging.basicConfig(format='%(message)s')

    parser = create_parser()
    # Implementation specific CLI options that are not needed to be part of the core parser.
    parser.add_argument('-a', '--atom', metavar='NAME', choices=['char', 'line', 'both'], default='line',
                        help='atom (i.e., granularity) of input (%(choices)s; default: %(default)s)')
    inators.arg.add_version_argument(parser, version=__version__)
    args = parser.parse_args()

    try:
        process_args(args)
    except ValueError as e:
        parser.error(e)

    out_src = reduce(args.src,
                     reduce_class=args.reduce_class,
                     reduce_config=args.reduce_config,
                     tester_class=args.tester_class,
                     tester_config=args.tester_config,
                     atom=args.atom,
                     cache_class=args.cache)

    postprocess(args, out_src)



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import codecs
import os
import shutil

from subprocess import run

from .outcome import Outcome


class SubprocessTest(object):

    def __init__(self, *, test_builder, command_pattern, work_dir, filename, encoding='utf-8', cleanup=True):
        """
        Wrapper around the script provided by the user. It decides about the
        interestingness based on the return code of executed script.

        :param test_builder: Callable object that creates test case from a
            configuration.
        :param command_pattern: The tester command as a sequence of arguments.
            If an element of the sequence contains %s, it is substituted with
            the path to the test case.
        :param work_dir: The work directory where test cases can be saved.
        :param filename: The file name to use for the test case.
        :param encoding: The encoding that will be used to save the tests.
        :param cleanup: Binary flag denoting whether the test directory should
            be removed after test execution or not.
        """
        self.test_builder = test_builder
        self.command_pattern = command_pattern
        self.work_dir = work_dir
        self.filename = filename
        self.encoding = encoding
        self.cleanup = cleanup

    def __call__(self, config, config_id):
        """
        Saving and evaluating of the current configuration.

        :param config: The list of units (chars or lines) that have to be
            compiled into a single test.
        :param config_id: Unique ID of the current configuration. It's used to
            name the containing folder of the current test.
        :return: The evaluation of the current test. It's either FAIL or PASS.
        """

        test_dir = os.path.join(self.work_dir, '_'.join(str(i) for i in config_id))
        test_path = os.path.join(test_dir, self.filename)

        os.makedirs(test_dir, exist_ok=True)

        with codecs.open(test_path, 'w', encoding=self.encoding, errors='ignore') as f:
            f.write(self.test_builder(config))

        args = []
        for arg in self.command_pattern:
            try:
                arg = arg % test_path
            except TypeError:
                pass
            args.append(arg)
        returncode = run(args, cwd=test_dir, check=False).returncode

        if self.cleanup:
            shutil.rmtree(test_dir)

        # Determine outcome.
        return Outcome.FAIL if returncode == 0 else Outcome.PASS


class ConcatTestBuilder(object):
    """
    Callable class that builds test case from a configuration.
    """

    def __init__(self, content):
        """
        Initialize a test builder with the atoms (e.g. chars or lines) of the
        original test case.

        :param content: Atoms of the original test case.
        """
        self._content = content

    def __call__(self, config):
        """
        Builds test case from the given config.

        :param config: Configuration to build a test case from.
        :return: Test case described by the config.
        """
        return ''.join(self._content[x] for x in config)
